##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for 0D heat exchanger models.

Author: John Eslick
"""
import pytest

from pyomo.environ import (ConcreteModel,
                           Constraint,
                           Expression,
                           TerminationCondition,
                           SolverStatus,
                           value,
                           Var)
from pyomo.common.config import ConfigBlock

from idaes.core import (FlowsheetBlock,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType)
from idaes.unit_models.heat_exchanger import (delta_temperature_lmtd_callback,
                                              HeatExchanger,
                                              HeatExchangerFlowPattern)

from idaes.property_models.activity_coeff_models.BTX_activity_coeff_VLE \
    import BTXParameterBlock
from idaes.property_models import iapws95
from idaes.property_models.examples.saponification_thermo import \
    SaponificationParameterBlock

from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_unused_variables)
from idaes.core.util.testing import (get_default_solver,
                                     PhysicalParameterTestBlock)
from pyomo.util.calc_var_value import calculate_variable_from_constraint


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()

# -----------------------------------------------------------------------------
def test_bad_option():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    with pytest.raises(KeyError):
        m.fs.unit = HeatExchanger(default={"I'm a bad option":"hot"})

def test_same_name():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    with pytest.raises(NameError):
        m.fs.unit = HeatExchanger(default={"cold_side_name":"shell"})

def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = HeatExchanger(default={
        "shell": {"property_package": m.fs.properties},
        "tube": {"property_package": m.fs.properties}})

    # Check unit config arguments
    # There are 8 to 10 arguments since you can add a side 1 and 2 config by
    # side_1, side_2, or whatever the user named them
    assert len(m.fs.unit.config) >= 8 and len(m.fs.unit.config) <= 10

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert isinstance(m.fs.unit.config.shell, ConfigBlock)
    assert isinstance(m.fs.unit.config.tube, ConfigBlock)
    assert m.fs.unit.config.delta_temperature_callback is \
        delta_temperature_lmtd_callback
    assert m.fs.unit.config.flow_pattern == \
        HeatExchangerFlowPattern.countercurrent

    # Check shell config
    assert len(m.fs.unit.config.shell) == 7
    assert m.fs.unit.config.shell.material_balance_type == \
        MaterialBalanceType.useDefault
    assert m.fs.unit.config.shell.energy_balance_type == \
        EnergyBalanceType.useDefault
    assert m.fs.unit.config.shell.momentum_balance_type == \
        MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.shell.has_phase_equilibrium
    assert not m.fs.unit.config.shell.has_pressure_change
    assert m.fs.unit.config.shell.property_package is m.fs.properties

    # Check tube config
    assert len(m.fs.unit.config.tube) == 7
    assert m.fs.unit.config.tube.material_balance_type == \
        MaterialBalanceType.useDefault
    assert m.fs.unit.config.tube.energy_balance_type == \
        EnergyBalanceType.useDefault
    assert m.fs.unit.config.tube.momentum_balance_type == \
        MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.tube.has_phase_equilibrium
    assert not m.fs.unit.config.tube.has_pressure_change
    assert m.fs.unit.config.tube.property_package is m.fs.properties

@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_costing():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = iapws95.Iapws95ParameterBlock()

    m.fs.unit = HeatExchanger(default={
                "shell": {"property_package": m.fs.properties},
                "tube": {"property_package": m.fs.properties},
                "flow_pattern": HeatExchangerFlowPattern.countercurrent})
#   Set inputs
    m.fs.unit.inlet_1.flow_mol[0].fix(100)
    m.fs.unit.inlet_1.enth_mol[0].fix(4000)
    m.fs.unit.inlet_1.pressure[0].fix(101325)

    m.fs.unit.inlet_2.flow_mol[0].fix(100)
    m.fs.unit.inlet_2.enth_mol[0].fix(3500)
    m.fs.unit.inlet_2.pressure[0].fix(101325)

    m.fs.unit.area.fix(1000)
    m.fs.unit.overall_heat_transfer_coefficient.fix(100)

    assert degrees_of_freedom(m) == 0

    m.fs.unit.initialize()

    m.fs.unit.get_costing()
    calculate_variable_from_constraint(
                m.fs.unit.costing.base_cost,
                m.fs.unit.costing.base_cost_eq)

    calculate_variable_from_constraint(
                m.fs.unit.costing.purchase_cost,
                m.fs.unit.costing.cp_cost_eq)

    results = solver.solve(m)
    assert m.fs.unit.costing.purchase_cost.value == \
                                            pytest.approx(52442.7363,1e-5)

# -----------------------------------------------------------------------------
class TestBTX_cocurrent(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = BTXParameterBlock(default={"valid_phase": 'Liq'})

        m.fs.unit = HeatExchanger(default={
                "shell": {"property_package": m.fs.properties},
                "tube": {"property_package": m.fs.properties},
                "flow_pattern": HeatExchangerFlowPattern.cocurrent})

        return m

    @pytest.mark.build
    def test_build(self, btx):
        assert hasattr(btx.fs.unit, "inlet_1")
        assert len(btx.fs.unit.inlet_1.vars) == 4
        assert hasattr(btx.fs.unit.inlet_1, "flow_mol")
        assert hasattr(btx.fs.unit.inlet_1, "mole_frac_comp")
        assert hasattr(btx.fs.unit.inlet_1, "temperature")
        assert hasattr(btx.fs.unit.inlet_1, "pressure")

        assert hasattr(btx.fs.unit, "inlet_2")
        assert len(btx.fs.unit.inlet_2.vars) == 4
        assert hasattr(btx.fs.unit.inlet_2, "flow_mol")
        assert hasattr(btx.fs.unit.inlet_2, "mole_frac_comp")
        assert hasattr(btx.fs.unit.inlet_2, "temperature")
        assert hasattr(btx.fs.unit.inlet_2, "pressure")

        assert hasattr(btx.fs.unit, "outlet_1")
        assert len(btx.fs.unit.outlet_1.vars) == 4
        assert hasattr(btx.fs.unit.outlet_1, "flow_mol")
        assert hasattr(btx.fs.unit.outlet_1, "mole_frac_comp")
        assert hasattr(btx.fs.unit.outlet_1, "temperature")
        assert hasattr(btx.fs.unit.outlet_1, "pressure")

        assert hasattr(btx.fs.unit, "outlet_2")
        assert len(btx.fs.unit.outlet_2.vars) == 4
        assert hasattr(btx.fs.unit.outlet_2, "flow_mol")
        assert hasattr(btx.fs.unit.outlet_2, "mole_frac_comp")
        assert hasattr(btx.fs.unit.outlet_2, "temperature")
        assert hasattr(btx.fs.unit.outlet_2, "pressure")

        assert isinstance(btx.fs.unit.overall_heat_transfer_coefficient, Var)
        assert isinstance(btx.fs.unit.area, Var)
        assert not hasattr(btx.fs.unit, "crossflow_factor")
        assert isinstance(btx.fs.unit.heat_duty, Var)
        assert isinstance(btx.fs.unit.delta_temperature_in, Var)
        assert isinstance(btx.fs.unit.delta_temperature_out, Var)
        assert isinstance(btx.fs.unit.unit_heat_balance, Constraint)
        assert isinstance(btx.fs.unit.delta_temperature, (Var, Expression))
        assert isinstance(btx.fs.unit.heat_transfer_equation, Constraint)

        assert number_variables(btx) == 50
        assert number_total_constraints(btx) == 38
        assert number_unused_variables(btx) == 0

    def test_dof(self, btx):
        btx.fs.unit.inlet_1.flow_mol[0].fix(5)  # mol/s
        btx.fs.unit.inlet_1.temperature[0].fix(365)  # K
        btx.fs.unit.inlet_1.pressure[0].fix(101325)  # Pa
        btx.fs.unit.inlet_1.mole_frac_comp[0, "benzene"].fix(0.5)
        btx.fs.unit.inlet_1.mole_frac_comp[0, "toluene"].fix(0.5)

        btx.fs.unit.inlet_2.flow_mol[0].fix(1)  # mol/s
        btx.fs.unit.inlet_2.temperature[0].fix(300)  # K
        btx.fs.unit.inlet_2.pressure[0].fix(101325)  # Pa
        btx.fs.unit.inlet_2.mole_frac_comp[0, "benzene"].fix(0.5)
        btx.fs.unit.inlet_2.mole_frac_comp[0, "toluene"].fix(0.5)

        btx.fs.unit.area.fix(1)
        btx.fs.unit.overall_heat_transfer_coefficient.fix(100)

        assert degrees_of_freedom(btx) == 0

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, btx):
        orig_fixed_vars = fixed_variables_set(btx)
        orig_act_consts = activated_constraints_set(btx)

        btx.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(btx) == 0

        fin_fixed_vars = fixed_variables_set(btx)
        fin_act_consts = activated_constraints_set(btx)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, btx):
        assert (pytest.approx(5, abs=1e-3) ==
                value(btx.fs.unit.outlet_1.flow_mol[0]))
        assert (pytest.approx(359.5, abs=1e-1) ==
                value(btx.fs.unit.outlet_1.temperature[0]))
        assert (pytest.approx(101325, abs=1e-3) ==
                value(btx.fs.unit.outlet_1.pressure[0]))

        assert (pytest.approx(1, abs=1e-3) ==
                value(btx.fs.unit.outlet_2.flow_mol[0]))
        assert (pytest.approx(329.9, abs=1e-1) ==
                value(btx.fs.unit.outlet_2.temperature[0]))
        assert (pytest.approx(101325, abs=1e-3) ==
                value(btx.fs.unit.outlet_2.pressure[0]))

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, btx):
        assert abs(value(btx.fs.unit.inlet_1.flow_mol[0] -
                         btx.fs.unit.outlet_1.flow_mol[0])) <= 1e-6
        assert abs(value(btx.fs.unit.inlet_2.flow_mol[0] -
                         btx.fs.unit.outlet_2.flow_mol[0])) <= 1e-6

        shell = value(
                btx.fs.unit.outlet_1.flow_mol[0] *
                (btx.fs.unit.shell.properties_in[0].enth_mol_phase['Liq'] -
                 btx.fs.unit.shell.properties_out[0].enth_mol_phase['Liq']))
        tube = value(
                btx.fs.unit.outlet_2.flow_mol[0] *
                (btx.fs.unit.tube.properties_in[0].enth_mol_phase['Liq'] -
                 btx.fs.unit.tube.properties_out[0].enth_mol_phase['Liq']))
        assert abs(shell + tube) <= 1e-6

    @pytest.mark.ui
    def test_report(self, btx):
        btx.fs.unit.report()

# -----------------------------------------------------------------------------
class TestBTX_cocurrent_alt_name(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = BTXParameterBlock(default={"valid_phase": 'Liq'})

        m.fs.unit = HeatExchanger(default={
                "hot_side_name":"hot",
                "cold_side_name":"cold",
                "hot": {"property_package": m.fs.properties},
                "cold": {"property_package": m.fs.properties},
                "flow_pattern": HeatExchangerFlowPattern.cocurrent})

        return m

    @pytest.mark.build
    def test_build(self, btx):
        assert hasattr(btx.fs.unit, "hot_inlet")
        assert len(btx.fs.unit.hot_inlet.vars) == 4
        assert hasattr(btx.fs.unit.hot_inlet, "flow_mol")
        assert hasattr(btx.fs.unit.hot_inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.hot_inlet, "temperature")
        assert hasattr(btx.fs.unit.hot_inlet, "pressure")

        assert hasattr(btx.fs.unit, "cold_inlet")
        assert len(btx.fs.unit.cold_inlet.vars) == 4
        assert hasattr(btx.fs.unit.cold_inlet, "flow_mol")
        assert hasattr(btx.fs.unit.cold_inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.cold_inlet, "temperature")
        assert hasattr(btx.fs.unit.cold_inlet, "pressure")

        assert hasattr(btx.fs.unit, "hot_outlet")
        assert len(btx.fs.unit.hot_outlet.vars) == 4
        assert hasattr(btx.fs.unit.hot_outlet, "flow_mol")
        assert hasattr(btx.fs.unit.hot_outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.hot_outlet, "temperature")
        assert hasattr(btx.fs.unit.hot_outlet, "pressure")

        assert hasattr(btx.fs.unit, "cold_outlet")
        assert len(btx.fs.unit.cold_outlet.vars) == 4
        assert hasattr(btx.fs.unit.cold_outlet, "flow_mol")
        assert hasattr(btx.fs.unit.cold_outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.cold_outlet, "temperature")
        assert hasattr(btx.fs.unit.cold_outlet, "pressure")

        assert isinstance(btx.fs.unit.overall_heat_transfer_coefficient, Var)
        assert isinstance(btx.fs.unit.area, Var)
        assert not hasattr(btx.fs.unit, "crossflow_factor")
        assert isinstance(btx.fs.unit.heat_duty, Var)
        assert isinstance(btx.fs.unit.delta_temperature_in, Var)
        assert isinstance(btx.fs.unit.delta_temperature_out, Var)
        assert isinstance(btx.fs.unit.unit_heat_balance, Constraint)
        assert isinstance(btx.fs.unit.delta_temperature, (Var, Expression))
        assert isinstance(btx.fs.unit.heat_transfer_equation, Constraint)

        assert number_variables(btx) == 50
        assert number_total_constraints(btx) == 38
        assert number_unused_variables(btx) == 0

    def test_dof(self, btx):
        btx.fs.unit.hot_inlet.flow_mol[0].fix(5)  # mol/s
        btx.fs.unit.hot_inlet.temperature[0].fix(365)  # K
        btx.fs.unit.hot_inlet.pressure[0].fix(101325)  # Pa
        btx.fs.unit.hot_inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        btx.fs.unit.hot_inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        btx.fs.unit.cold_inlet.flow_mol[0].fix(1)  # mol/s
        btx.fs.unit.cold_inlet.temperature[0].fix(300)  # K
        btx.fs.unit.cold_inlet.pressure[0].fix(101325)  # Pa
        btx.fs.unit.cold_inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        btx.fs.unit.cold_inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        btx.fs.unit.area.fix(1)
        btx.fs.unit.overall_heat_transfer_coefficient.fix(100)

        assert degrees_of_freedom(btx) == 0

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, btx):
        orig_fixed_vars = fixed_variables_set(btx)
        orig_act_consts = activated_constraints_set(btx)

        btx.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(btx) == 0

        fin_fixed_vars = fixed_variables_set(btx)
        fin_act_consts = activated_constraints_set(btx)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, btx):
        assert (pytest.approx(5, abs=1e-3) ==
                value(btx.fs.unit.hot_outlet.flow_mol[0]))
        assert (pytest.approx(359.5, abs=1e-1) ==
                value(btx.fs.unit.hot_outlet.temperature[0]))
        assert (pytest.approx(101325, abs=1e-3) ==
                value(btx.fs.unit.hot_outlet.pressure[0]))

        assert (pytest.approx(1, abs=1e-3) ==
                value(btx.fs.unit.cold_outlet.flow_mol[0]))
        assert (pytest.approx(329.9, abs=1e-1) ==
                value(btx.fs.unit.cold_outlet.temperature[0]))
        assert (pytest.approx(101325, abs=1e-3) ==
                value(btx.fs.unit.cold_outlet.pressure[0]))

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, btx):
        assert abs(value(btx.fs.unit.hot_inlet.flow_mol[0] -
                         btx.fs.unit.hot_outlet.flow_mol[0])) <= 1e-6
        assert abs(value(btx.fs.unit.cold_inlet.flow_mol[0] -
                         btx.fs.unit.cold_outlet.flow_mol[0])) <= 1e-6

        shell = value(
                btx.fs.unit.hot_outlet.flow_mol[0] *
                (btx.fs.unit.hot.properties_in[0].enth_mol_phase['Liq'] -
                 btx.fs.unit.hot.properties_out[0].enth_mol_phase['Liq']))
        tube = value(
                btx.fs.unit.cold_outlet.flow_mol[0] *
                (btx.fs.unit.cold.properties_in[0].enth_mol_phase['Liq'] -
                 btx.fs.unit.cold.properties_out[0].enth_mol_phase['Liq']))
        assert abs(shell + tube) <= 1e-6

    @pytest.mark.ui
    def test_report(self, btx):
        btx.fs.unit.report()


# -----------------------------------------------------------------------------
@pytest.mark.iapws
@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
class TestIAPWS_countercurrent(object):
    @pytest.fixture(scope="class")
    def iapws(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = iapws95.Iapws95ParameterBlock()

        m.fs.unit = HeatExchanger(default={
                "shell": {"property_package": m.fs.properties},
                "tube": {"property_package": m.fs.properties},
                "flow_pattern": HeatExchangerFlowPattern.countercurrent})

        return m

    @pytest.mark.build
    def test_build(self, iapws):
        assert len(iapws.fs.unit.inlet_1.vars) == 3
        assert hasattr(iapws.fs.unit.inlet_1, "flow_mol")
        assert hasattr(iapws.fs.unit.inlet_1, "enth_mol")
        assert hasattr(iapws.fs.unit.inlet_1, "pressure")

        assert hasattr(iapws.fs.unit, "outlet_1")
        assert len(iapws.fs.unit.outlet_1.vars) == 3
        assert hasattr(iapws.fs.unit.outlet_1, "flow_mol")
        assert hasattr(iapws.fs.unit.outlet_1, "enth_mol")
        assert hasattr(iapws.fs.unit.outlet_1, "pressure")

        assert len(iapws.fs.unit.inlet_2.vars) == 3
        assert hasattr(iapws.fs.unit.inlet_2, "flow_mol")
        assert hasattr(iapws.fs.unit.inlet_2, "enth_mol")
        assert hasattr(iapws.fs.unit.inlet_2, "pressure")

        assert hasattr(iapws.fs.unit, "outlet_2")
        assert len(iapws.fs.unit.outlet_2.vars) == 3
        assert hasattr(iapws.fs.unit.outlet_2, "flow_mol")
        assert hasattr(iapws.fs.unit.outlet_2, "enth_mol")
        assert hasattr(iapws.fs.unit.outlet_2, "pressure")

        assert isinstance(iapws.fs.unit.overall_heat_transfer_coefficient, Var)
        assert isinstance(iapws.fs.unit.area, Var)
        assert not hasattr(iapws.fs.unit, "crossflow_factor")
        assert isinstance(iapws.fs.unit.heat_duty, Var)
        assert isinstance(iapws.fs.unit.delta_temperature_in, Var)
        assert isinstance(iapws.fs.unit.delta_temperature_out, Var)
        assert isinstance(iapws.fs.unit.unit_heat_balance, Constraint)
        assert isinstance(iapws.fs.unit.delta_temperature, (Expression, Var))
        assert isinstance(iapws.fs.unit.heat_transfer_equation, Constraint)

        assert number_variables(iapws) == 18
        assert number_total_constraints(iapws) == 10
        assert number_unused_variables(iapws) == 0

    def test_dof(self, iapws):
        iapws.fs.unit.inlet_1.flow_mol[0].fix(100)
        iapws.fs.unit.inlet_1.enth_mol[0].fix(4000)
        iapws.fs.unit.inlet_1.pressure[0].fix(101325)

        iapws.fs.unit.inlet_2.flow_mol[0].fix(100)
        iapws.fs.unit.inlet_2.enth_mol[0].fix(3500)
        iapws.fs.unit.inlet_2.pressure[0].fix(101325)

        iapws.fs.unit.area.fix(1000)
        iapws.fs.unit.overall_heat_transfer_coefficient.fix(100)

        assert degrees_of_freedom(iapws) == 0

    def test_dof_alt_name1(self, iapws):
        iapws.fs.unit.shell_inlet.flow_mol[0].fix(100)
        iapws.fs.unit.shell_inlet.enth_mol[0].fix(4000)
        iapws.fs.unit.shell_inlet.pressure[0].fix(101325)

        iapws.fs.unit.tube_inlet.flow_mol[0].fix(100)
        iapws.fs.unit.tube_inlet.enth_mol[0].fix(3500)
        iapws.fs.unit.tube_inlet.pressure[0].fix(101325)

        iapws.fs.unit.area.fix(1000)
        iapws.fs.unit.overall_heat_transfer_coefficient.fix(100)

        assert degrees_of_freedom(iapws) == 0

    @pytest.mark.initialization
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, iapws):
        orig_fixed_vars = fixed_variables_set(iapws)
        orig_act_consts = activated_constraints_set(iapws)

        iapws.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(iapws) == 0

        fin_fixed_vars = fixed_variables_set(iapws)
        fin_act_consts = activated_constraints_set(iapws)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, iapws):
        results = solver.solve(iapws)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, iapws):
        assert pytest.approx(100, abs=1e-5) == \
            value(iapws.fs.unit.outlet_1.flow_mol[0])
        assert pytest.approx(100, abs=1e-5) == \
            value(iapws.fs.unit.outlet_2.flow_mol[0])

        assert pytest.approx(3535, abs=1e0) == \
            value(iapws.fs.unit.outlet_1.enth_mol[0])
        assert pytest.approx(3964.5, abs=1e0) == \
            value(iapws.fs.unit.outlet_2.enth_mol[0])

        assert pytest.approx(101325, abs=1e2) == \
            value(iapws.fs.unit.outlet_1.pressure[0])
        assert pytest.approx(101325, abs=1e2) == \
            value(iapws.fs.unit.outlet_2.pressure[0])

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, iapws):
        assert abs(value(iapws.fs.unit.inlet_1.flow_mol[0] -
                         iapws.fs.unit.outlet_1.flow_mol[0])) <= 1e-6
        assert abs(value(iapws.fs.unit.inlet_2.flow_mol[0] -
                         iapws.fs.unit.outlet_2.flow_mol[0])) <= 1e-6

        shell_side = value(
                iapws.fs.unit.outlet_1.flow_mol[0] *
                (iapws.fs.unit.inlet_1.enth_mol[0] -
                 iapws.fs.unit.outlet_1.enth_mol[0]))
        tube_side = value(
                iapws.fs.unit.outlet_2.flow_mol[0] *
                (iapws.fs.unit.inlet_2.enth_mol[0] -
                 iapws.fs.unit.outlet_2.enth_mol[0]))
        assert abs(shell_side + tube_side) <= 1e-6

    @pytest.mark.ui
    def test_report(self, iapws):
        iapws.fs.unit.report()


# -----------------------------------------------------------------------------
#@pytest.mark.skip(reason="Solutions vary with differnt versions of solver.")
class TestSaponification_crossflow(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = SaponificationParameterBlock()

        m.fs.unit = HeatExchanger(default={
                "shell": {"property_package": m.fs.properties},
                "tube": {"property_package": m.fs.properties},
                "flow_pattern": HeatExchangerFlowPattern.crossflow})

        return m

    @pytest.mark.build
    def test_build(self, sapon):
        assert len(sapon.fs.unit.inlet_1.vars) == 4
        assert hasattr(sapon.fs.unit.inlet_1, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet_1, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet_1, "temperature")
        assert hasattr(sapon.fs.unit.inlet_1, "pressure")

        assert len(sapon.fs.unit.outlet_1.vars) == 4
        assert hasattr(sapon.fs.unit.outlet_1, "flow_vol")
        assert hasattr(sapon.fs.unit.outlet_1, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.outlet_1, "temperature")
        assert hasattr(sapon.fs.unit.outlet_1, "pressure")

        assert len(sapon.fs.unit.inlet_2.vars) == 4
        assert hasattr(sapon.fs.unit.inlet_2, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet_2, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet_2, "temperature")
        assert hasattr(sapon.fs.unit.inlet_2, "pressure")

        assert len(sapon.fs.unit.outlet_2.vars) == 4
        assert hasattr(sapon.fs.unit.outlet_2, "flow_vol")
        assert hasattr(sapon.fs.unit.outlet_2, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.outlet_2, "temperature")
        assert hasattr(sapon.fs.unit.outlet_2, "pressure")

        assert isinstance(sapon.fs.unit.overall_heat_transfer_coefficient, Var)
        assert isinstance(sapon.fs.unit.area, Var)
        assert isinstance(sapon.fs.unit.crossflow_factor, Var)
        assert isinstance(sapon.fs.unit.heat_duty, Var)
        assert isinstance(sapon.fs.unit.delta_temperature_in, Var)
        assert isinstance(sapon.fs.unit.delta_temperature_out, Var)
        assert isinstance(sapon.fs.unit.unit_heat_balance, Constraint)
        assert isinstance(sapon.fs.unit.delta_temperature, (Expression,Var))
        assert isinstance(sapon.fs.unit.heat_transfer_equation, Constraint)

        assert number_variables(sapon) == 39
        assert number_total_constraints(sapon) == 20
        assert number_unused_variables(sapon) == 0

    def test_dof(self, sapon):
        sapon.fs.unit.inlet_1.flow_vol[0].fix(1e-3)
        sapon.fs.unit.inlet_1.temperature[0].fix(320)
        sapon.fs.unit.inlet_1.pressure[0].fix(101325)
        sapon.fs.unit.inlet_1.conc_mol_comp[0, "H2O"].fix(55388.0)
        sapon.fs.unit.inlet_1.conc_mol_comp[0, "NaOH"].fix(100.0)
        sapon.fs.unit.inlet_1.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        sapon.fs.unit.inlet_1.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        sapon.fs.unit.inlet_1.conc_mol_comp[0, "Ethanol"].fix(0.0)

        sapon.fs.unit.inlet_2.flow_vol[0].fix(1e-3)
        sapon.fs.unit.inlet_2.temperature[0].fix(300)
        sapon.fs.unit.inlet_2.pressure[0].fix(101325)
        sapon.fs.unit.inlet_2.conc_mol_comp[0, "H2O"].fix(55388.0)
        sapon.fs.unit.inlet_2.conc_mol_comp[0, "NaOH"].fix(100.0)
        sapon.fs.unit.inlet_2.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        sapon.fs.unit.inlet_2.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        sapon.fs.unit.inlet_2.conc_mol_comp[0, "Ethanol"].fix(0.0)

        sapon.fs.unit.area.fix(1000)
        sapon.fs.unit.overall_heat_transfer_coefficient.fix(100)
        sapon.fs.unit.crossflow_factor.fix(0.6)

        assert degrees_of_freedom(sapon) == 0

    @pytest.mark.initialization
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, sapon):
        orig_fixed_vars = fixed_variables_set(sapon)
        orig_act_consts = activated_constraints_set(sapon)

        sapon.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(sapon) == 0

        fin_fixed_vars = fixed_variables_set(sapon)
        fin_act_consts = activated_constraints_set(sapon)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, sapon):
        results = solver.solve(sapon)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, sapon):
        assert pytest.approx(1e-3, abs=1e-6) == \
            value(sapon.fs.unit.outlet_1.flow_vol[0])
        assert pytest.approx(1e-3, abs=1e-6) == \
            value(sapon.fs.unit.outlet_2.flow_vol[0])

        assert pytest.approx(55388.0, rel=1e-3) == value(
                sapon.fs.unit.outlet_1.conc_mol_comp[0, "H2O"])
        assert pytest.approx(100.0, rel=1e-3) == value(
                sapon.fs.unit.outlet_1.conc_mol_comp[0, "NaOH"])
        assert pytest.approx(100.0, rel=1e-3) == value(
                sapon.fs.unit.outlet_1.conc_mol_comp[0, "EthylAcetate"])
        assert pytest.approx(0.0, abs=1e-3) == value(
                sapon.fs.unit.outlet_1.conc_mol_comp[0, "SodiumAcetate"])
        assert pytest.approx(0.0, abs=1e-3) == value(
                sapon.fs.unit.outlet_1.conc_mol_comp[0, "Ethanol"])

        assert pytest.approx(55388.0, rel=1e-3) == value(
                sapon.fs.unit.outlet_2.conc_mol_comp[0, "H2O"])
        assert pytest.approx(100.0, rel=1e-3) == value(
                sapon.fs.unit.outlet_2.conc_mol_comp[0, "NaOH"])
        assert pytest.approx(100.0, rel=1e-3) == value(
                sapon.fs.unit.outlet_2.conc_mol_comp[0, "EthylAcetate"])
        assert pytest.approx(0.0, abs=1e-3) == value(
                sapon.fs.unit.outlet_2.conc_mol_comp[0, "SodiumAcetate"])
        assert pytest.approx(0.0, abs=1e-3) == value(
                sapon.fs.unit.outlet_2.conc_mol_comp[0, "Ethanol"])

        assert pytest.approx(301.3, abs=1e-1) == \
            value(sapon.fs.unit.outlet_1.temperature[0])
        assert pytest.approx(318.7, abs=1e-1) == \
            value(sapon.fs.unit.outlet_2.temperature[0])

        assert pytest.approx(101325, abs=1e2) == \
            value(sapon.fs.unit.outlet_1.pressure[0])
        assert pytest.approx(101325, abs=1e2) == \
            value(sapon.fs.unit.outlet_2.pressure[0])

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, sapon):
        shell_side = value(
                sapon.fs.unit.outlet_1.flow_vol[0] *
                sapon.fs.properties.dens_mol*sapon.fs.properties.cp_mol *
                (sapon.fs.unit.inlet_1.temperature[0] -
                 sapon.fs.unit.outlet_1.temperature[0]))
        tube_side = value(
                sapon.fs.unit.outlet_2.flow_vol[0] *
                sapon.fs.properties.dens_mol*sapon.fs.properties.cp_mol *
                (sapon.fs.unit.inlet_2.temperature[0] -
                 sapon.fs.unit.outlet_2.temperature[0]))
        assert abs(shell_side + tube_side) <= 1e0

    @pytest.mark.ui
    def test_report(self, sapon):
        sapon.fs.unit.report()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This module contains the base class for constructing flowsheet models in the
IDAES modeling framework.
"""

import logging

import pyomo.environ as pe
from pyomo.dae import ContinuousSet
from pyomo.network import Arc
from pyomo.common.config import ConfigValue, In

from idaes.core import (ProcessBlockData, declare_process_block_class,
                        UnitModelBlockData, useDefault)
from idaes.core.util.config import (is_physical_parameter_block,
                                    is_time_domain,
                                    list_of_floats)
from idaes.core.util.exceptions import DynamicError
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.dmf.ui.flowsheet_serializer import FlowsheetSerializer

from idaes.core.util import unit_costing as costing

# Some more information about this module
__author__ = "John Eslick, Qi Chen, Andrew Lee"


__all__ = ['FlowsheetBlock', 'FlowsheetBlockData']

# Set up logger
_log = logging.getLogger(__name__)


@declare_process_block_class("FlowsheetBlock", doc="""
    FlowsheetBlock is a specialized Pyomo block for IDAES flowsheet models, and
    contains instances of FlowsheetBlockData.""")
class FlowsheetBlockData(ProcessBlockData):
    """
    The FlowsheetBlockData Class forms the base class for all IDAES process
    flowsheet models. The main purpose of this class is to automate the tasks
    common to all flowsheet models and ensure that the necessary attributes of
    a flowsheet model are present.

    The most signfiicant role of the FlowsheetBlockData class is to
    automatically create the time domain for the flowsheet.
    """

    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare("dynamic", ConfigValue(
        default=useDefault,
        domain=In([useDefault, True, False]),
        description="Dynamic model flag",
        doc="""Indicates whether this model will be dynamic,
**default** - useDefault.
**Valid values:** {
**useDefault** - get flag from parent or False,
**True** - set as a dynamic model,
**False** - set as a steady-state model.}"""))
    CONFIG.declare("time", ConfigValue(
        default=None,
        domain=is_time_domain,
        description="Flowsheet time domain",
        doc="""Pointer to the time domain for the flowsheet. Users may provide
an existing time domain from another flowsheet, otherwise the flowsheet will
search for a parent with a time domain or create a new time domain and
reference it here."""))
    CONFIG.declare("time_set", ConfigValue(
        default=[0],
        domain=list_of_floats,
        description="Set of points for initializing time domain",
        doc="""Set of points for initializing time domain. This should be a
list of floating point numbers,
**default** - [0]."""))
    CONFIG.declare("default_property_package", ConfigValue(
        default=None,
        domain=is_physical_parameter_block,
        description="Default property package to use in flowsheet",
        doc="""Indicates the default property package to be used by models
within this flowsheet if not otherwise specified,
**default** - None.
**Valid values:** {
**None** - no default property package,
**a ParameterBlock object**.}"""))

    def build(self):
        """
        General build method for FlowsheetBlockData. This method calls a number
        of sub-methods which automate the construction of expected attributes
        of flowsheets.

        Inheriting models should call `super().build`.

        Args:
            None

        Returns:
            None
        """
        super(FlowsheetBlockData, self).build()

        # Set up dynamic flag and time domain
        self._setup_dynamics()

    def is_flowsheet(self):
        """
        Method which returns True to indicate that this component is a
        flowsheet.

        Args:
            None

        Returns:
            True
        """
        return True

    def model_check(self):
        """
        This method runs model checks on all unit models in a flowsheet.

        This method searches for objects which inherit from UnitModelBlockData
        and executes the model_check method if it exists.

        Args:
            None

        Returns:
            None
        """
        _log.info("Executing model checks.")
        for o in self.component_objects(descend_into=False):
            if isinstance(o, UnitModelBlockData):
                try:
                    o.model_check()
                except AttributeError:
                    # This should never happen, but just in case
                    _log.warning('{} Model/block has no model_check method.'
                                 .format(o.name))

    def stream_table(self, true_state=False, time_point=0, orient='columns'):
        """
        Method to generate a stream table by iterating over all Arcs in the
        flowsheet.

        Args:
            true_state : whether the state variables (True) or display
                         variables (False, default) from the StateBlocks should
                         be used in the stream table.
            time_point : point in the time domain at which to create stream
                         table (default = 0)
            orient : whether stream should be shown by columns ("columns") or
                     rows ("index")

        Returns:
            A pandas dataframe containing stream table information
        """
        dict_arcs = {}

        for a in self.component_objects(ctype=Arc, descend_into=False):
            dict_arcs[a.local_name] = a

        return create_stream_table_dataframe(dict_arcs,
                                             time_point=time_point,
                                             orient=orient,
                                             true_state=true_state)

    def serialize(self, file_base_name, overwrite=False):
        """
        Serializes the flowsheet and saves it to a file that can be read by the
        idaes-model-vis  jupyter lab extension.

        :param file_base_name: The file prefix to the .idaes.vis file produced.
        The file is created/saved
        in the directory that you ran from Jupyter Lab.
        :param overwrite: Boolean to overwrite an existing file_base_name.idaes.vis.
        If True, the existing file with the same file_base_name will be overwritten.
        This will cause you to lose any saved layout.
        If False and there is an existing file with that file_base_name, you will get
        an error message stating that you cannot save a file to the file_base_name
        (and therefore overwriting the saved layout). If there is not an existing
        file with that file_base_name then it saves as normal.
        Defaults to False.
        :return: None
        """
        serializer = FlowsheetSerializer()
        serializer.serialize(self, file_base_name, overwrite)

    def get_costing(self, module=costing, year=None):
        self.costing = pe.Block()

        module.global_costing_parameters(self.costing, year)

    def _get_stream_table_contents(self, time_point=0):
        """
        Calls stream_table method and returns result
        """
        return self.stream_table(time_point)

    def _setup_dynamics(self):
        # Look for parent flowsheet
        fs = self.flowsheet()

        # Check the dynamic flag, and retrieve if necessary
        if self.config.dynamic == useDefault:
            if fs is None:
                # No parent, so default to steady-state and warn user
                _log.warning('{} is a top level flowsheet, but dynamic flag '
                             'set to useDefault. Dynamic '
                             'flag set to False by default'
                             .format(self.name))
                self.config.dynamic = False

            else:
                # Get dynamic flag from parent flowsheet
                self.config.dynamic = fs.config.dynamic

        # Check for case when dynamic=True, but parent dynamic=False
        elif self.config.dynamic is True:
            if fs is not None and fs.config.dynamic is False:
                raise DynamicError(
                        '{} trying to declare a dynamic model within '
                        'a steady-state flowsheet. This is not '
                        'supported by the IDAES framework. Try '
                        'creating a dynamic flowsheet instead, and '
                        'declaring some models as steady-state.'
                        .format(self.name))

        if self.config.time is not None:
            # Validate user provided time domain
            if (self.config.dynamic is True and
                    not isinstance(self.config.time, ContinuousSet)):
                raise DynamicError(
                        '{} was set as a dynamic flowsheet, but time domain '
                        'provided was not a ContinuousSet.'.format(self.name))
        else:
            # If no parent flowsheet, set up time domain
            if fs is None:
                # Create time domain
                if self.config.dynamic:
                    # Check if time_set has at least two points
                    if len(self.config.time_set) < 2:
                        # Check if time_set is at default value
                        if self.config.time_set == [0.0]:
                            # If default, set default end point to be 1.0
                            self.config.time_set = [0.0, 1.0]
                        else:
                            # Invalid user input, raise Excpetion
                            raise DynamicError(
                                    "Flowsheet provided with invalid "
                                    "time_set attribute - must have at "
                                    "least two values (start and end).")
                    # For dynamics, need a ContinuousSet
                    self.time = ContinuousSet(initialize=self.config.time_set)
                else:
                    # For steady-state, use an ordered Set
                    self.time = pe.Set(initialize=self.config.time_set,
                                       ordered=True)

                # Set time config argument as reference to time domain
                self.config.time = self.time
            else:
                # Set time config argument to parent time
                self.config.time = fs.config.time



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Jupyter magics for the DMF.
"""
# stdlib
import inspect
import logging
import os
import webbrowser
# third-party
import pendulum
from IPython.core.magic import Magics, magics_class, line_magic
from IPython.display import display_markdown
# local
from . import dmfbase, errors, help, workspace

__author__ = 'Dan Gunter <dkgunter@lbl.gov>'

# Logging

_log = logging.getLogger(__name__)


# Custom exception classes

class DMFMagicError(errors.DMFError):
    def __init__(self, errmsg, usermsg=None):
        super(DMFMagicError, self).__init__(errmsg)
        if usermsg is True:
            usermsg = errmsg
        self.message = usermsg


@magics_class
class DmfMagics(Magics):
    """Implement "magic" commands in Jupyter/IPython for
    interacting with the DMF and IDAES more generally.

    In order to allow easier testing, the functionality is
    broken into two classes. This class has the decorated method(s)
    for invoking the 'magics', and :class:`DmfMagicsImpl` has the
    state and functionality.
    """

    def __init__(self, shell):
        super(DmfMagics, self).__init__(shell)
        self._impl = DmfMagicsImpl(shell)
        self._last_ok = None

    @property
    def last_ok(self):
        return self._last_ok is not False  # None or True

    @line_magic
    def dmf(self, line):
        """DMF outer command.

        Example::

            %dmf <subcommand> [subcommand args..]
        """
        try:
            result = self._impl.dmf(line)
            self._last_ok = True
        except DMFMagicError as err:
            _log.error('Error with magic command: {}'.format(err))
            result = None  # err.message
        return result


class DmfMagicsImpl(object):
    """State and implementation called by DmfMagics.

    On failure of any method, a `DMFMagicError` is raised, that
    should be handled by the line or cell magic that invoked it.
    """
    # Encoding which commands need 'init'.
    # Also encode args required to that command.
    # Key is command, value is #args: '*'=any, '+'=1 or more.
    _NEED_INIT = {
        'info': '*',
        'help': '+'
    }

    def __init__(self, shell):
        self._dmf = None
        self._shell = shell

    @property
    def initialized(self):
        return self._dmf is not None

    def dmf(self, line):
        """DMF outer command
        """
        # Parse input into a subcommand and other tokens
        line = line.strip()
        if line == '':
            # No command will invoke "help"
            tokens = ['help']
        else:
            tokens = line.split()
        # Find sub-method matching subcommand
        subcmd = tokens[0]
        submeth = getattr(self, 'dmf_' + subcmd, None)
        if submeth is None:
            txt = "Unrecognized command `{}`".format(subcmd)
            # self._dmf_markdown(txt)
            raise DMFMagicError(txt)
        # Invoke sub-method
        params = tokens[1:]
        try:
            return submeth(*params)
        except Exception as err:
            msg = 'Error in `%dmf {}`: {}'.format(subcmd, err)
            # self._dmf_markdown(msg)
            raise DMFMagicError(msg, usermsg=msg)

    def dmf_init(self, path, *extra):
        """Initialize DMF (do this before most other commands).
        *Arguments*: `path ["create"]`

        Args:
            path (str): Full path to DMF home
            extra (str): Extra tokens. If 'create', then try to create the
                         path if it is not found.
        """
        kwargs, create = {}, False
        if len(extra) > 0:
            if extra[0].lower() == 'create':
                if os.path.exists(
                        os.path.join(path, dmfbase.DMF.WORKSPACE_CONFIG)):
                    self._dmf_markdown(
                        'Cannot create new configuration at "{}": '
                        'file "{}" exists. Will try without '
                        '"create" option'
                        .format(path, dmfbase.DMF.WORKSPACE_CONFIG))
                else:
                    kwargs = {'create': True, 'add_defaults': True}
                    create = True
            else:
                _log.warning('Ignoring extra argument to "init"')
        try:
            self._dmf = dmfbase.DMF(path, **kwargs)
        except errors.WorkspaceNotFoundError:
            assert not create
            msg = 'Workspace not found at path "{}". ' \
                  'If you want to create a new workspace, add the word ' \
                  '"create", after the path, to the command.'.format(path)
            # self._dmf_markdown(msg)
            raise DMFMagicError(msg)
        except errors.WorkspaceCannotCreateError as err:
            txt = 'Workspace could not be created at path "{}": {}' \
                .format(path, err)
            raise DMFMagicError(txt, usermsg=txt)
        except errors.DMFError as err:
            raise DMFMagicError('Initializing workspace: {}'.format(err),
                                usermsg=True)

        self._dmf.set_meta({'name': os.path.basename(path)})
        if create:
            self._dmf_markdown('*Success!* Created new workspace at "{}"'
                               .format(path))
        else:
            self._dmf_markdown('*Success!* Using workspace at "{}"'
                               .format(path))

    def dmf_workspaces(self, *paths):
        """List DMF workspaces. Optionally takes one or more paths
        to use as a starting point. By default, start from current
        directory. *Arguments*: `[paths..]`

        Args:
            paths (List[str]): Paths to search, use "." by default
        """
        wslist = []
        if len(paths) == 0:
            paths = ['.']
        for root in paths:
            wslist.extend(workspace.find_workspaces(root))
        any_good_workspaces = False
        output_table = []
        for w in sorted(wslist):
            try:
                ws = workspace.Workspace(w)
                output_table.append((w, ws.name, ws.description))
                any_good_workspaces = True
            except errors.WorkspaceError:
                pass  # XXX: Should we print a warning?
        if not any_good_workspaces:
            # either no paths, or all paths raised an error
            return 'No valid workspaces found\n'
        else:
            lines = ['| Path | Name | Description |',
                     '| ---- | ---- | ----------- |']
            for row in output_table:
                rowstr = '| {} | {} | {} |'.format(row[0], row[1], row[2])
                lines.append(rowstr)
            listing = '\n'.join(lines)
            self._dmf_markdown(listing)
            return '{:d} workspaces found\n'.format(len(listing))

    def dmf_list(self):
        """List resources in the current workspace. *Arguments*: none.
        """
        self._init_required('list')
        lines = ['| ID | Name(s) | Type | Modified | Description | ',
                 '| -- | ------- | ---- | -------- | ----------- |']
        for rsrc in self._dmf.find():
            msince = pendulum.from_timestamp(rsrc.v['modified']) \
                .diff_for_humans()
            rowstr = '| {id} | {names} | {type} | {mdate} | {desc} |'.format(
                id=rsrc.id, names=','.join(rsrc.v['aliases']), type=rsrc.type,
                mdate=msince, desc=rsrc.v['desc'])
            lines.append(rowstr)
        listing = '\n'.join(lines)
        self._dmf_markdown(listing)
        return True

    def dmf_info(self, *topics):
        """Provide information about DMF current state. *Arguments*: none

        Args:
            (List[str]) topics: List of topics

        Returns:
            None
        """
        self._init_required('info')
        if topics:
            self._dmf_markdown('Sorry, no topic-specific info yet available')
            raise DMFMagicError('Topics not supported')
        # configuration info
        text_lines = ['## Configuration']
        for key, value in self._dmf.meta.items():
            hdr = '  * {}'.format(key)
            if isinstance(value, list):
                text_lines.append('{}:'.format(hdr))
                for v in value:
                    text_lines.append('    - {}'.format(v))
            elif isinstance(value, dict):
                text_lines.append('{}:'.format(hdr))
                for k2, v2 in value.items():
                    text_lines.append('    - {}: {}'.format(k2, v2))
            else:
                text_lines.append('{}: {}'.format(hdr, value))
        conf_info = '\n'.join(text_lines)
        # all info
        all_info = '\n'.join((conf_info,))
        self._dmf_markdown(all_info)

    def dmf_help(self, *names):
        """Provide help on IDAES objects and classes.
        Invoking with no arguments gives general help.
        Invoking with one argument looks for help in the docs
        on the given object or class.
        *Arguments*: `[name]`.
        """
        if len(names) == 0:
            # give some general help for magics
            return self._magics_help()
        self._init_required('help')
        if len(names) > 1:
            _log.warning('DMF Help is restricted to only one object or class '
                         'at a time. Ignoring trailing arguments ({})'
                         .format(' '.join(names[1:])))
        name = names[0]
        # Check some special names first.
        # To re-use the <module>.<class> mechanism, translate them into
        # some pseudo-classes in the "dmf.help" pseudo-module.
        p_module, p_class = None, None
        if name.lower() == 'dmf':
            p_module = 'dmf'
            p_class = 'DMF'
        elif name.lower() in ('help',):
            p_module = 'dmf'
            p_class = 'Help'
        elif name.lower() in ('idaes', '*', ''):
            p_module = 'idaes'
            p_class = 'Home'
        # Get the help
        helpfiles = None
        if p_module is not None:
            try:
                helpfiles = help.get_html_docs(self._dmf, p_module, p_class)
            except DMFMagicError as err:
                _log.debug('Getting help for pseudo-class {}::{}, error: {}'
                           .format(p_module, p_class, err))
                # for Result
                name = 'pseudo-module {}.{}'.format(p_module, p_class)
        else:
            try:
                helpfiles = self._find_help_for_object(name)
            except DMFMagicError as err:
                _log.debug('Getting help for object {}, error: {}'
                           .format(name, err))
        # Result
        if helpfiles:
            self._show_help_in_browser(helpfiles)
        else:
            self._dmf_markdown('No Sphinx docs found for "{}"'.format(name))
        return None

    def _init_required(self, subcmd):
        """If no DMF, display init-required message and
        raise an exception. Otherwise, do nothing and return False.
        """
        if self._dmf is not None:
            return False
        msg = 'Must call `init` before command `{}`'.format(subcmd)
        self._dmf_markdown(msg)
        raise DMFMagicError(msg)

    @staticmethod
    def _dmf_markdown(text):
        display_markdown(text, raw=True)

    def _find_help_for_object(self, name):
        """Get object by evaluating name as an expression.
        If that doesn't work, try just using it as a string.
        """
        obj, oname = None, None
        try:
            obj = self._shell.ev(name)
            _log.debug('Looking for HTML docs for object: {}'.format(obj))
        except Exception:
            oname = name
            _log.debug('Looking for HTML docs for object: {}'.format(oname))
        try:
            result = help.find_html_docs(self._dmf, obj=obj, obj_name=oname)
        except (ValueError, AttributeError):
            if obj is not None:
                raise DMFMagicError('Cannot find help for {}'.format(obj))
            else:
                raise DMFMagicError('Cannot find help for {}'.format(oname))
        return result

    @staticmethod
    def _show_help_in_browser(helpfiles):
        """Open help docs in the browser."""
        for hf in helpfiles:
            url = 'file://' + hf
            _log.debug('Opening URL "{}"'.format(url))
            webbrowser.open_new(url)

    def _magics_help(self):
        """Introspect to give a list of commands."""
        # Build a dictionary of magic methods and their
        # descriptions (from their docstrings).
        help_dict = {'dmf': {}, 'idaes': {}}
        for name, meth in inspect.getmembers(self, inspect.ismethod):
            if name.startswith('dmf_'):
                help_dict['dmf'][name] = self._extract_help_text(meth)
            elif name.startswith('idaes_'):
                help_dict['idaes'][name] = self._extract_help_text(meth)
        # build help text from dict
        txt = ''
        for mname in 'idaes', 'dmf':
            if len(help_dict[mname]) == 0:
                continue
            txt_list = []
            for name in sorted(help_dict[mname].keys()):
                cmd = name[len(mname) + 1:]
                cmd_help = '-  `%{} {}` - {}'.format(
                    mname, cmd, help_dict[mname][name])
                txt_list.append(cmd_help)
            section = '{}## {} magic commands'.format(
                '\n\n' if txt else '', mname.upper())
            txt += section + '\n\n' + '\n'.join(txt_list)
        return self._dmf_markdown(txt)

    @staticmethod
    def _extract_help_text(meth):
        doc = inspect.getdoc(meth)
        # extract first sentence (strip off period)
        sentence = doc[:doc.find('\n\n')].strip('.')
        # take out line breaks from sentence
        hdr = sentence.replace('\n', ' ')
        return hdr


###############################################################################
# Register at import time
###############################################################################

_registered = False


def register():
    """Register with IPython on import (once).
    """
    global _registered
    if _registered:
        return
    try:
        ip = get_ipython()  # noqa: F821
        _log.debug('Registering DMF magics')
        ip.register_magics(DmfMagics)
    except:  # noqa: E722
        pass
    _registered = True


register()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for feed block.
Authors: Andrew Lee
"""

import pytest
from pyomo.environ import (ConcreteModel,
                           TerminationCondition,
                           SolverStatus,
                           value)
from idaes.core import FlowsheetBlock
from idaes.unit_models.feed import Feed
from idaes.property_models.examples.saponification_thermo import (
                        SaponificationParameterBlock)
from idaes.property_models import iapws95
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_unused_variables)
from idaes.core.util.testing import (get_default_solver,
                                     PhysicalParameterTestBlock)


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()


# -----------------------------------------------------------------------------
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = Feed(default={"property_package": m.fs.properties})

    # Check unit config arguments
    assert len(m.fs.unit.config) == 4

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.property_package is m.fs.properties


# -----------------------------------------------------------------------------
class TestSaponification(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = SaponificationParameterBlock()

        m.fs.unit = Feed(default={"property_package": m.fs.properties})

        return m

    @pytest.mark.build
    def test_build(self, sapon):

        assert hasattr(sapon.fs.unit, "flow_vol")
        assert hasattr(sapon.fs.unit, "conc_mol_comp")
        assert hasattr(sapon.fs.unit, "temperature")
        assert hasattr(sapon.fs.unit, "pressure")

        assert hasattr(sapon.fs.unit, "outlet")
        assert len(sapon.fs.unit.outlet.vars) == 4
        assert hasattr(sapon.fs.unit.outlet, "flow_vol")
        assert hasattr(sapon.fs.unit.outlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.outlet, "temperature")
        assert hasattr(sapon.fs.unit.outlet, "pressure")

        assert number_variables(sapon) == 8
        assert number_total_constraints(sapon) == 0
        assert number_unused_variables(sapon) == 8

    def test_dof(self, sapon):
        sapon.fs.unit.flow_vol.fix(1.0e-03)
        sapon.fs.unit.conc_mol_comp[0, "H2O"].fix(55388.0)
        sapon.fs.unit.conc_mol_comp[0, "NaOH"].fix(100.0)
        sapon.fs.unit.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        sapon.fs.unit.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        sapon.fs.unit.conc_mol_comp[0, "Ethanol"].fix(0.0)

        sapon.fs.unit.temperature.fix(303.15)
        sapon.fs.unit.pressure.fix(101325.0)

        assert degrees_of_freedom(sapon) == 0

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, sapon):
        orig_fixed_vars = fixed_variables_set(sapon)
        orig_act_consts = activated_constraints_set(sapon)

        sapon.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(sapon) == 0

        fin_fixed_vars = fixed_variables_set(sapon)
        fin_act_consts = activated_constraints_set(sapon)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    # No solve, as nothing to solve for

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, sapon):
        assert (pytest.approx(101325.0, abs=1e-2) ==
                value(sapon.fs.unit.outlet.pressure[0]))
        assert (pytest.approx(303.15, abs=1e-2) ==
                value(sapon.fs.unit.outlet.temperature[0]))
        assert (pytest.approx(1e-3, abs=1e-5) ==
                value(sapon.fs.unit.outlet.flow_vol[0]))
        assert (pytest.approx(55388, abs=1e0) ==
                value(sapon.fs.unit.outlet.conc_mol_comp[0, "H2O"]))
        assert (pytest.approx(100.0, abs=1e-2) ==
                value(sapon.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]))
        assert (pytest.approx(100.0, abs=1e-2) ==
                value(sapon.fs.unit.outlet.conc_mol_comp[0, "NaOH"]))
        assert (pytest.approx(0.00, abs=1e-2) ==
                value(sapon.fs.unit.outlet.conc_mol_comp[0, "Ethanol"]))
        assert (pytest.approx(0.00, abs=1e-2) ==
                value(sapon.fs.unit.outlet.conc_mol_comp[0, "SodiumAcetate"]))

    @pytest.mark.ui
    def test_report(self, sapon):
        sapon.fs.unit.report()


# -----------------------------------------------------------------------------
@pytest.mark.iapws
@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
class TestIAPWS(object):
    @pytest.fixture(scope="class")
    def iapws(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = iapws95.Iapws95ParameterBlock(default={
                "phase_presentation": iapws95.PhaseType.LG})

        m.fs.unit = Feed(default={"property_package": m.fs.properties})

        return m

    @pytest.mark.build
    def test_build(self, iapws):
        assert hasattr(iapws.fs.unit, "flow_mol")
        assert hasattr(iapws.fs.unit, "enth_mol")
        assert hasattr(iapws.fs.unit, "pressure")

        assert hasattr(iapws.fs.unit, "outlet")
        assert len(iapws.fs.unit.outlet.vars) == 3
        assert hasattr(iapws.fs.unit.outlet, "flow_mol")
        assert hasattr(iapws.fs.unit.outlet, "enth_mol")
        assert hasattr(iapws.fs.unit.outlet, "pressure")

        assert number_variables(iapws) == 3
        assert number_total_constraints(iapws) == 0
        assert number_unused_variables(iapws) == 3

    def test_dof(self, iapws):
        iapws.fs.unit.flow_mol.fix(100)
        iapws.fs.unit.enth_mol.fix(24000)
        iapws.fs.unit.pressure.fix(101325)

        assert degrees_of_freedom(iapws) == 0

    @pytest.mark.initialization
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, iapws):
        orig_fixed_vars = fixed_variables_set(iapws)
        orig_act_consts = activated_constraints_set(iapws)

        iapws.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(iapws) == 0

        fin_fixed_vars = fixed_variables_set(iapws)
        fin_act_consts = activated_constraints_set(iapws)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    # No solve test, as nothing to solve for

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, iapws):
        assert (pytest.approx(101325.0, abs=1e3) ==
                value(iapws.fs.unit.outlet.pressure[0]))
        assert (pytest.approx(24000, abs=1e3) ==
                value(iapws.fs.unit.outlet.enth_mol[0]))
        assert (pytest.approx(100.0, abs=1e-2) ==
                value(iapws.fs.unit.outlet.flow_mol[0]))

        assert (pytest.approx(373.12, abs=1e-2) == value(
            iapws.fs.unit.properties[0].temperature))
        assert (pytest.approx(0.5953, abs=1e-4) == value(
            iapws.fs.unit.properties[0].phase_frac["Liq"]))

    @pytest.mark.ui
    def test_report(self, iapws):
        iapws.fs.unit.report()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tabular data handling
"""
# standard
import abc
import csv
import json
import logging
import re
# third-party
import jsonschema
try:
    import pandas as pd
    import numpy as np
except ImportError:
    np, pd = None, None
# local
from . import errors
from .util import get_file

__author__ = 'Dan Gunter <dkgunter@lbl.gov>'

_log = logging.getLogger(__name__)


class Fields(object):
    """Constants for field names.
    """
    DATA, META = 'data', 'meta'
    DTYPE, AUTH, INFO, TITLE, DATE = \
        'datatype', 'authors', 'info', 'title', 'date'
    VALS, ROWS = 'values', 'rows'
    #: Keys for data mapping
    DATA_NAME = 'name'
    DATA_UNITS = 'units'
    DATA_VALUES = 'values'
    DATA_ERRORS = 'errors'
    DATA_ERRTYPE = 'error_type'
    # Used during parsing
    COLTYPE = 'type'

# --------------------------------------------------------------------------
# Schemas


DATA_SCHEMA_DEF = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "examples": [
                "Density",
                "r"
            ]
        },
        "units": {
            "type": "string",
            "examples": [
                "mPa-s",
                "K"
            ]
        },
        "values": {
            "description": "Column of numeric values",
            "type": "array",
            "items": {
                "type": "number"
            },
            "examples": ["[2.6, 6.21]"]
        },
        "error_type": {
            "description": "Type for error values",
            "type": "string"
        },
        "errors": {
            "description": "Column of numeric errors",
            "type": "array",
            "items": {
                "type": "number"
            },
            "examples": [
                "[0.001, 0.035]"
            ]
        },
        "type": {
            "description": "Type of column",
            "enum": ["state", "property"]
        }
    },
    "required": [
        "name",
        "units",
        "values"
    ],
    "additionalProperties": False
}

METADATA_SCHEMA_DEF = {
    "type": "object",
    "properties": {
        "datatype": {
            "description": "name of the data type",
            "type": "string",
            "examples": ["MEA"]
        },
        "info": {
            "description": "Additional information about the source "
                           "(i.e. publication)",
            "type": "string",
            "examples": [
                "J. Chem. Eng. Data, 2009, Vol 54, pg. 3096-30100"]
        },
        "notes": {
            "description": "Free-form text with notes about the data",
            "type": "string",
            "examples": ["r is MEA weight fraction in aqueous soln."]
        },
        "authors": {
            "description": "Author list in format Last1, First1, Last2,"
                           " First2, etc.",
            "type": "string",
            "examples": ["Amundsen, T.G., Lars, E.O., Eimer, D.A."]
        },
        "title": {
            "description": "Title of the source (e.g. publication"
                           " title)",
            "type": "string",
            "examples": [
                "Density and Viscosity of Monoethanolamine + .etc."]
        },
        "date": {
            "description": "Date of source data",
            "type": "string",
            "examples": ["2009"]
        }
    },
    "required": [
        "datatype",
        "authors",
        "title",
        "date"
    ],
    "additionalProperties": True
}

TABLE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "id": "http://idaes.org/table",
    "type": "object",
    "properties": {
        "meta": {
            "description": "List of information about the data source",
            "type": "array",
            "items": METADATA_SCHEMA_DEF
        },
        "data": {
            "description": "Measured data columns",
            "type": "array",
            "items": DATA_SCHEMA_DEF
        }
    },
    "required": ["meta", "data"],
    "additionalProperties": False
}

COLUMN_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "id": "http://idaes.org/table",
    "type": "array",
    "items": DATA_SCHEMA_DEF,
}

# Schemas
# --------------------------------------------------------------------------


class TabularObject(object):
    """Abstract Property data class.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def as_dict(self):
        """Return Python dict representation.
        """
        pass


class Table(TabularObject):
    """Tabular data and metadata together (at last!)
    """
    _validator = jsonschema.Draft4Validator(TABLE_SCHEMA)

    def __init__(self, data=None, metadata=None):
        """Wrapper object for data + metadata of properties.

        Args:
            data (list|TabularData): Raw data dictionaries
            metadata (dict|list|Metadata): Metadata dictionar(ies)
        """
        if isinstance(data, list):
            self._data = TabularData(data)
        elif isinstance(data, TabularData):
            self._data = data
        else:
            raise TypeError('Expected list or TabularData object for: {}'
                            .format(data))
        self._meta = []
        if metadata:
            if isinstance(metadata, list) or isinstance(metadata, tuple):
                for m in metadata:
                    self.add_metadata(m)
            else:
                self.add_metadata(metadata)

    def __iter__(self):
        yield 'data', self._data.as_list()
        yield 'meta', [m.as_dict() for m in self._meta]

    def add_metadata(self, m):
        if isinstance(m, dict):
            obj = Metadata(values=m)
        elif isinstance(m, Metadata):
            obj = m
        else:
            raise TypeError('Expected dict or Metadata object for: {}'
                            .format(m))
        self._meta.append(obj)

    @property
    def data(self):
        return self._data

    @property
    def metadata(self):
        return self._meta

    def as_dict(self):
        """Represent as a Python dictionary.

        Returns:
            (dict) Dictionary representation
        """
        return {k: v for k, v in self}

    def dump(self, fp, **kwargs):
        """Dump to file as JSON.
        Convenience method, equivalent to converting to a
        dict and calling :meth:`json.dump`.

        Args:
            fp (file): Write output to this file
            **kwargs: Keywords passed to json.dump()

        Returns:
            see json.dump()
        """
        return json.dump(self.as_dict(), fp, **kwargs)

    def dumps(self, **kwargs):
        """Dump to string as JSON.
        Convenience method, equivalent to converting to a
        dict and calling :meth:`json.dumps`.

        Args:
            **kwargs: Keywords passed to json.dumps()

        Returns:
            (str) JSON-formatted data
        """
        return json.dumps(self.as_dict(), **kwargs)

    def __str__(self):
        return self.dumps()

    @classmethod
    def load(cls, file_or_path, validate=True):
        """Create from JSON input.

        Args:
            file_or_path (file or str): Filename or file object
                from which to read the JSON-formatted data.
            validate (bool): If true, apply validation to input JSON data.

        Example input::

            {
                "meta": [{
                    "datatype": "MEA",
                    "info": "J. Chem. Eng. Data, 2009, Vol 54, pg. 3096-30100",
                    "notes": "r is MEA weight fraction in aqueous soln.",
                    "authors": "Amundsen, T.G., Lars, E.O., Eimer, D.A.",
                    "title": "Density and Viscosity of Monoethanolamine + etc."
                }],
                "data": [
                    {
                        "name": "Viscosity Value",
                        "units": "mPa-s",
                        "values": [2.6, 6.2],
                        "error_type": "absolute",
                        "errors": [0.06, 0.004],
                        "type": "property"
                    }
                ]
            }
        """
        fp = get_file(file_or_path)
        d = json.load(fp)
        if validate:
            cls._validate_json(d)
        metalist = d[Fields.META]
        meta = [Metadata(m) for m in metalist]
        data = TabularData(d[Fields.DATA])
        tbl = Table(data=data)
        for m in meta:
            tbl.add_metadata(m)
        return tbl

    @classmethod
    def _validate_json(cls, d):
        # print('@@ validating:\n----\n{}\n-----'.format(d))
        try:
            cls._validator.validate(d)
        except jsonschema.ValidationError as err:
            raise errors.DataFormatError('tabulardata', str(err))


class TabularData(object):
    """Class representing tabular data that knows how to
    construct itself from a CSV file.

    You can build objects from multiple CSV files as well.
    See the property database section of the API docs for
    details, or read the code in :meth:`add_csv` and the
    tests in :mod:`idaes_dmf.propdb.tests.test_mergecsv`.
    """
    embedded_units = r'(.*)\((.*)\)'

    _validator = jsonschema.Draft4Validator(COLUMN_SCHEMA)

    def __init__(self, data, error_column=False):
        """Construct from a list.

            [  {
                "name": "Density Data",
                "units": "g/cm^3",
                "values": [1.0053, 1.0188, .., ],
                "errors": [.00005, .., .00005],
                "error_type": "absolute"
              },
              ...etc...
            ]

        Args:
            data (list): Input dictionary
            error_column (bool): Whether there are error columns
        Returns:
            TabularData: New instance.
        Raises:
            TypeError: Bad type for `data`
            ValueError: Bad value for `data`
        """
        if not isinstance(data, list):
            raise TypeError('Expected list of dicts, got {}'
                            .format(type(data)))
        if len(data) == 0:
            raise ValueError('Input data must have at least one column')
        try:
            self._validator.validate(data)
        except jsonschema.ValidationError as err:
            raise ValueError(str(err))
        self._data = data
        self._nrows = self._get_nrows()
        self._errcol = error_column

    @property
    def columns(self):
        return self._data

    def __len__(self):
        return self._nrows

    def names(self):
        """Get column names.

        Returns:
            list[str]: List of column names.
        """
        return [v[Fields.DATA_NAME] for v in self.columns]

    @property
    def num_columns(self):
        """Number of columns in this table.

        A "column" is defined as data + error. So if there
        are two columns of data, each with an associated
        error column, then `num_columns` is 2 (not 4).

        Returns:
            int: Number of columns.
        """
        return len(self.columns)

    @property
    def num_rows(self):
        """Number of rows in this table.

        obj.num_rows is a synonym for len(obj)

        Returns:
            int: Number of rows.
        """
        return self._nrows

    def _get_nrows(self):
        n = 0
        for v in self.columns:
            vals = v[Fields.DATA_VALUES]
            if n == 0:
                n = len(vals)
            elif len(vals) != n:
                raise ValueError('Column "{}" length {} != {}'
                                 .format(v[Fields.DATA_NAME], len(vals), n))
        return n

    def get_column(self, key):
        """Get an object for the given named column.

        Args:
            key (str): Name of column

        Returns:
            (TabularColumn) Column object.

        Raises:
            KeyError: No column by that name.
        """
        result = None
        for v in self.columns:
            if v[Fields.DATA_NAME] == key:
                result = Column(key, v)
                break
        if result is None:
            name_list = ', '.join(self.names())
            raise KeyError('Bad column name "{}", not in ({})'.format(
                key, name_list))
        return result

    def get_column_index(self, key):
        """Get an index for the given named column.

        Args:
            key (str): Name of column

        Returns:
            (int) Column number.

        Raises:
            KeyError: No column by that name.
        """
        # print('@@ get column index for name: {}'.format(key))
        for i, v in enumerate(self.columns):
            if v[Fields.DATA_NAME] == key:
                return i
        raise KeyError('Bad column name "{}", not in ({})'.format(
            key, ', '.join(self.names())))

    def as_list(self):
        """Export the data as a list.

        Output will be in same form as data passed to constructor.

        Returns:
            (list) List of dicts
        """
        return self._data

    def as_arr(self):
        """Export property data as arrays.

        Returns:
            (values[M,N], errors[M,N]) Two arrays of floats,
            each with M columns having N values.

        Raises:
            ValueError if the columns are not all the same length
        """
        values, errvals = [], []
        # extract columns
        for v in self.columns:
            values.append(v[Fields.DATA_VALUES])
            errvals.append(v[Fields.DATA_ERRORS])
        return values, errvals

    def values_dataframe(self):
        """Get values as a dataframe.

        Returns:
            (pd.DataFrame) Pandas dataframe for values.

        Raises:
            ImportError: If `pandas` or `numpy` were never
                successfully imported.
        """
        return self._get_dataframe(Fields.DATA_VALUES)

    def errors_dataframe(self):
        """Get errors as a dataframe.

        Returns:
            pd.DataFrame: Pandas dataframe for values.

        Raises:
            ImportError: If `pandas` or `numpy` were never
                successfully imported.
        """
        return self._get_dataframe(Fields.DATA_ERRORS)

    def _get_dataframe(self, field):
        self._check_pandas_import()
        a1, names = [], []
        a1.extend([v[field] for v in self.columns])
        names.extend([v[Fields.DATA_NAME] for v in self.columns])
        a2 = np.array(a1).transpose()
        return pd.DataFrame(a2, columns=names)

    @staticmethod
    def _check_pandas_import():
        if pd is None:
            raise ImportError('Failed to import Pandas and/or Numpy packages '
                              'at module load. Cannot return a Pandas '
                              'Dataframe without Pandas.')

    @staticmethod
    def from_csv(file_or_path, error_column=False):
        """Import the CSV data.

        Expected format of the  files is a header plus data rows.

        Header: Index-column,  Column-name(1), Error-column(1),  \
                Column-name(2), Error-column(2), ..
        Data: <index>, <val>, <errval>, <val>, <errval>, ..

        Column-name is in the format "Name (units)"

        Error-column is in the format "<type> Error", where "<type>" is
        the error type.

        Args:
            file_or_path (file-like or str): Input file
            error_column (bool): If True, look for an error column after each
                           value column. Otherwise, all columns are
                           assumed to be values.

        Returns:
            TabularData: New table of data
        """
        input_file = get_file(file_or_path)
        csv_file = csv.reader(input_file)
        row = next(csv_file)
        names, data = TabularData._parse_csv_headers(row,
                                                     error_column=error_column)
        for row in csv_file:
            # print('@@ parse csv row: {}'.format(row))
            TabularData._parse_csv_row(data, row, error_column=error_column)
        obj = TabularData(data, error_column=error_column)
        return obj

    @classmethod
    def _parse_csv_headers(cls, headers, error_column=None):
        """Parse a row of CSV headers which are pairs
        of columns like "<name> [(units)], <error-type> Error".

        Returns:
             (names, data). Names is a list of all the column names.
                            Data is a list of property/state objects.
        """
        if error_column:
            if len(headers) < 3:
                raise ValueError('Less than 3 columns')
            if len(headers) % 2 != 1:
                raise ValueError('Number of columns must be odd')
        else:
            if len(headers) < 2:
                raise ValueError('Less than 2 columns')
        data = []
        all_names = []
        # Add new item for each value/error pair in column headers
        column_step = 2 if error_column else 1
        for i in range(1, len(headers), column_step):
            errhdr = ''
            if error_column:
                hdr, errhdr = headers[i], headers[i + 1]
            else:
                hdr = headers[i]
            m = re.match(cls.embedded_units, hdr)
            name, units = m.groups() if m else (hdr, '')
            name = name.strip()  # ignore extra ws for column names
            if error_column:
                errtype = errhdr.strip().split()[0].lower()
            else:
                errtype = 'none'
            item = {Fields.DATA_NAME: name,
                    Fields.DATA_UNITS: units,
                    Fields.DATA_VALUES: [],
                    Fields.DATA_ERRORS: [],
                    Fields.DATA_ERRTYPE: errtype}
            data.append(item)
            all_names.append(name)
        return all_names, data

    @classmethod
    def _parse_csv_row(cls, data, row, error_column=None):
        """Add data in row to dict in data, which has the form
        returned by `_parse_csv_headers`.

        Rows are laid out like this:

            id, state-value1, state-error1, state-value2, state-error2, ..., \
            prop-value1, prop-error1, prop-value2, prop-error2, ...

        The number of state-value/error column pairs is equal to `nstates`.
        """
        rowlen = len(row)
        # check that the row has the right number of columns
        if error_column:
            expected_rowlen = 2 * len(data) + 1
        else:
            expected_rowlen = len(data) + 1
        if rowlen != expected_rowlen:
            raise ValueError('CSV row, expected {:d} columns, got {:d}'
                             .format(expected_rowlen, rowlen))
        # iterate over the values in the row, adding each to
        # the appropriate values in data['states'] or data['properties']
        column_step = 2 if error_column else 1
        for i in range(1, rowlen, column_step):
            value = row[i]
            value = float(value) if value else float('nan')
            # add value and error to the property column
            c = data[(i - 1) // column_step]
            c[Fields.DATA_VALUES].append(value)
            if error_column:
                error = row[i + 1]
                error = float(error) if error else float('nan')
                c[Fields.DATA_ERRORS].append(error)


class Metadata(object):
    """Class to import metadata.
    """
    # generic line regular expression
    line_expr = re.compile(r'\s*(\w+)\s*,\s*(.*)\s*')
    # source line regular expression
    source_expr = re.compile(r'\s*(.*)\s*,\s*"(.*)"\s*,\s*(.*)\s*')

    def __init__(self, values=None):
        if values is None:
            values = {}
        self._meta = values

    @property
    def datatype(self):
        return self._meta.get(Fields.DTYPE, '')

    @datatype.setter
    def datatype(self, value):
        self._meta[Fields.DTYPE] = value

    @property
    def author(self):
        """Publication author(s)."""
        return self._meta[Fields.AUTH]

    @property
    def date(self):
        """Publication date"""
        return self._meta.get(Fields.DATE, '1970-01-01')

    @property
    def title(self):
        """Publication title."""
        return self._meta[Fields.TITLE]

    @property
    def info(self):
        """Publication venue, etc."""
        return self._meta[Fields.INFO]

    @property
    def source(self):
        """Full publication info."""
        return '{},"{}",{}'.format(self.author, self.title, self.info)

    def as_dict(self):
        return self._meta

    @staticmethod
    def from_csv(file_or_path):
        """Import metadata from simple text format.

        Example input::

            Source,Han, J., Jin, J., Eimer, D.A., Melaaen, M.C.,"Density of \
            Water(1) + Monoethanolamine(2) + CO2(3) from (298.15 to 413.15) K\
            and Surface Tension of Water(1) + Monethanolamine(2) from ( \
            303.15 to 333.15)K", J. Chem. Eng. Data, 2012, Vol. 57, \
            pg. 1095-1103"
            Retrieval,"J. Morgan, date unknown"
            Notes,r is MEA weight fraction in aqueous soln. (CO2-free basis)

        Args:
            file_or_path (str or file): Input file

        Returns:
            (PropertyMetadata) New instance
        """
        meta = {}
        f = get_file(file_or_path)
        got_source = False
        for line in f:
            line = line.strip()
            if len(line) == 0 or line[0] == '#':
                continue
            m = Metadata.line_expr.match(line)
            if m is None:
                raise ValueError('Cannot parse metadata from: {}'.format(line))
            label, info = m.groups()
            label = label.lower()
            if label == 'source':
                meta.update(Metadata._parse_source(info))
                got_source = True
            else:
                meta[label] = info
        if not got_source:
            raise ValueError('"source" is required')
        return Metadata(meta)

    @staticmethod
    def _parse_source(s):
        m = Metadata.source_expr.match(s)
        if m is None:
            raise ValueError('Cannot parse source: {}'.format(s))
        auth, title, rest = m.groups()
        date_match = re.search(r',\s*(\d\d\d\d(?:-\d\d(?:-\d\d)))', rest)
        date = date_match.group(1) if date_match else '1970-01-01'
        return {Fields.AUTH: auth,
                Fields.DATE: date,
                Fields.TITLE: title,
                Fields.INFO: rest}


class Column(object):
    """Generic, abstract column
    """
    type_name = 'generic'

    def __init__(self, name, data):
        self.name = name
        self.units = data[Fields.DATA_UNITS]
        self.values = data[Fields.DATA_VALUES]
        self.entity_type = self.type_name
        if Fields.DATA_ERRORS in data:
            self.errors = data[Fields.DATA_ERRORS]
            self.error_type = data[Fields.DATA_ERRTYPE]

    def data(self):
        return {
            Fields.DATA_UNITS: self.units,
            Fields.DATA_VALUES: self.values,
            Fields.DATA_ERRORS: self.errors,
            Fields.DATA_ERRTYPE: self.error_type
        }

    def __len__(self):
        return len(self.values)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
HELMholtz Energy Thermodynamics (HELMET)

Main capabilities of HELMET
default HELMET use
"""

__author__ = "Marissa Engle <mengle@andrew.cmu.edu>"

import os
import platform
import subprocess


from idaes.surrogate import alamopy
from . import AncillaryEquations  # , Certainty
from . import Plotting, DataImport, DataManipulation
from . import GAMSWrite, BasisFunctions

from matplotlib import cm


# global R
R = 8.314472  # kJ mol^-1 K^-1

# global molecule, filename, gamsname, data_name
molecule, filename, gamsname, data_name = None, None, None, None

# global sample, sample_ratio
sample = False
sample_ratio = 5

# global critT, critP, critD, M, triple, acc
critT, critP, critD, M, triple, acc = 0, 0, 0, 0, 0, 0

# global max_time, num_terms
max_time = 500
num_terms = 12

# global props
props = []

# global has_alamo
has_alamo = None

# global flag_dirty
flag_dirty = False


def initialize(**kwargs):
    """
    filename - location of data
    gamsname - name of the gams file made
    molecule - name of the molecule/compound
    data_name - name of the data 
    fluid data - [critT, critP, critD, M, triple, acentric factor] 
    R - gas constant value
    """

    global R, molecule, filename, gamsname, data_name
    global critT, critP, critD, M, triple, acc
    global max_time, num_terms, props
    global sample, sample_ratio
    global flag_dirty
    global has_alamo

    k_dict = {
        "R": R,
        "filename": filename,
        "gamsname": gamsname,
        "molecule": molecule,
        "fluid_data": (critT, critP, critD, M, triple, acc),
        "max_time": max_time,
        "num_terms": num_terms,
        "props": props,
        "sample": 1,
    }

    for arg in k_dict:
        if arg in kwargs:
            if arg == "R":
                R = kwargs[arg]
            elif arg == "filename":
                filename = kwargs[arg]
            elif arg == "gamsname":
                gamsname = kwargs[arg]
            elif arg == "molecule":
                molecule = kwargs[arg]
            elif arg == "fluid_data":
                (critT, critP, critD, M, triple, acc) = kwargs[arg]
            elif arg == "max_time":
                max_time = kwargs[arg]
            elif arg == "num_terms":
                num_terms = kwargs[arg]
            elif arg == "props":
                props = kwargs[arg]
            elif arg == "sample":
                sample = True
                sample_ratio = kwargs[arg]
            else:
                raise Exception("Not a keyword argument")

    if has_alamo is None:
        has_alamo = alamopy.multos.has_alamo()
        if not has_alamo:
            print("No ALAMO software found.")

    updateModelSettings()

    flag_dirty = True


def updateModelSettings():
    """
    Settings of the model based on the chemical passed to the
    different python methods
    """

    global R, molecule, filename, gamsname, data_name
    global critT, critP, critD, M, triple, acc
    global max_time, num_terms, props
    global flag_dirty

    # SoaveDensity.molData((critT, critP, critD, M, triple, acc), molecule, R)
    Plotting.molData((critT, critP, critD, M, triple, acc), molecule, R)
    Plotting.props = props
    DataImport.molData((critT, critP, critD, M, triple, acc), R)
    DataImport.filename = filename
    DataManipulation.molData((critT, critP, critD, M, triple, acc), molecule, R)
    AncillaryEquations.molecule = molecule
    AncillaryEquations.max_time = max_time
    GAMSWrite.molData(
        (critT, critP, critD, M, triple, acc), molecule, data_name, num_terms, max_time
    )
    BasisFunctions.molData((critT, critP, critD, M, triple, acc), molecule, R)

    flag_dirty = False


def prepareAncillaryEquations(plot=False, keepFiles=False):
    """
    Develops ancillary equations of state using ALAMOPY
        DL - saturated liquid density 
        DV - saturated vapor density
        PV - vapor pressure
    Dependent on ALAMO
    """
    global has_alamo

    if has_alamo:
        AncillaryEquations.DL()
        AncillaryEquations.DV()
        AncillaryEquations.PV()

        if plot:
            Plotting.viewAnc()

        if not keepFiles:
            for p in ["DL", "DV", "PV"]:
                os.remove("%s%s" % (molecule, p))
                os.remove("%s%s.lst" % (molecule, p))
    else:
        if plot:
            Plotting.viewAnc()
        print("Couldn't regress ancillary equations. ALAMO executable not found")



def viewPropertyData():
    """
    Plot imported data
    """
    Plotting.viewData()


def setupRegression(numTerms=14, gams=False, pyomo=False):
    """
    setup gams regression
    """
    global props, sample, sample_ratio
    GAMSWrite.num_terms = numTerms
    GAMSWrite.props = props
    GAMSWrite.sample = sample
    GAMSWrite.sample = sample_ratio

    GAMSWrite.importData()

    GAMSWrite.runFile = "gdx"
    GAMSWrite.GenerateGDXGamsFiledtlmv()

    GAMSWrite.runFile = "main"
    GAMSWrite.GenerateGamsShell()


def runRegression(gams = False, pyomo=False):
    """
    Runs the gdx and main regression gams file
    """

    GAMSWrite.runFile ="gdx"
    command = "gams %s%s.gms"%(molecule,GAMSWrite.runFile)
    process = subprocess.check_call(command, shell=True)

    GAMSWrite.runFile ="main"
    command = "gams %s%s.gms"%(molecule,GAMSWrite.runFile)
    process = subprocess.check_call(command, shell=True)


def getFlag():
    """
    Returns flag, marks a change in the construction of the model
    """
    return flag_dirty


def viewResults(lstFile=None, plot = False, report=False, surface=cm.coolwarm):
    """
    Plot results from gams or pyomo
        lstFile - gams listing file
        surface - colormapping color eg. cm.coolwarm
    """
    Plotting.sseCombo(lstFile = lstFile, plot= plot, report=report, surface= surface)


def viewMultResults(lstFile, numTerms=0):
    """
    View mutliple results from a lst file 
    """
    Plotting.multSSECombo(lstFile, numTerms)


def deletefile(*fname):
    """
    Deletes files
    """
    tos = platform.platform()
    if "Windows" in tos:
        for name in fname:
            os.system("del " + name)
    else:
        for name in fname:
            os.system("rm " + name)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

__author__ = "John Eslick"

"""Transformation to replace variables with other variables."""
import pyomo.environ as pyo
from pyomo.core.base.plugin import TransformationFactory
from pyomo.core.plugins.transform.hierarchy import NonIsomorphicTransformation
from pyomo.core.expr import current as EXPR
from pyomo.common.config import ConfigBlock, ConfigValue, add_docstring_list
from pyomo.core.base.var import _GeneralVarData


def _is_var(v):
    return isinstance(v, (_GeneralVarData, pyo.Var))

@TransformationFactory.register(
    'replace_variables',
    doc="Replace variables with other variables.")
class ReplaceVariables(NonIsomorphicTransformation):
    """Replace variables in a model or block with other variables.

    Keyword arguments below are specified for the ``apply_to(instance, **kwargs)``
    method.

    """
    CONFIG = ConfigBlock()
    CONFIG.declare("substitute", ConfigValue(
        default=[],
        description="List-like of tuples where the first item in a tuple is a "
                    "Pyomo variable to be replaced and the second item in the "
                    "tuple is a Pyomo variable to replace it with. This "
                    "transformation is not reversible."
    ))

    __doc__ = add_docstring_list(__doc__, CONFIG)

    @staticmethod
    def replace(instance, substitute):
        # Create the replacement dict. Do some argument validation and indexed
        # var handling
        d = {}
        for r in substitute:
            if not (_is_var(r[0]) and _is_var(r[1])):
                raise TypeError(
                    "Replace only allows variables to be replaced, {} is type {}"
                    " and {} is type {}".format(r[0], type(r[0]), r[1], type(r[1]))
                )
            if r[0].is_indexed() != r[1].is_indexed():
                raise TypeError(
                    "IndexedVars must be replaced by IndexedVars, {} is type {}"
                    " and {} is type {}".format(r[0], type(r[0]), r[1], type(r[1]))
                )
            if r[0].is_indexed() and r[1].is_indexed():
                if not r[0].index_set().issubset(r[1].index_set()):
                    raise ValueError(
                        "The index set of {} must be a subset of"
                        " {}.".format(r[0], r[1])
                    )
                for i in r[0]:
                    d[id(r[0][i])] = r[1][i]
            else:
                #scalar replace
                d[id(r[0])] = r[1]

        # Replacement Visitor
        vis = EXPR.ExpressionReplacementVisitor(substitute=d)

        # Do replacements in Expressions, Constraints, and Objectives
        for c in instance.component_data_objects(pyo.Constraint, descend_into=True):
            c.set_value(expr=vis.dfs_postorder_stack(c.expr))
        for c in instance.component_data_objects((pyo.Expression, pyo.Objective),
                                                 descend_into=True):
            vis.dfs_postorder_stack(c)


    def _apply_to(self, instance, **kwds):
        """
        Apply the transformation.  This is called by ``apply_to`` in the
        superclass, and should not be called directly.  ``apply_to`` takes the
        same arguments.

        Args:
            instance: A block or model to apply the transformation to
            substitute: A list-like of two-element list-likes.  Each two element
                list-like specifies a replacment of the first variable by the
                second.  SimpleVar, IndexedVar, _GeneralVarData, and Reference are
                all accepted types.

        Returns:
            None
        """
        config = self.CONFIG(kwds)
        self.replace(instance, config.substitute)



import pyomo.common.config
import logging.config
import toml
import os
import importlib


_log = logging.getLogger(__name__)

default_config = """
default_binary_url = "https://github.com/IDAES/idaes-ext/releases/download/1.0.1/"
use_idaes_solvers = true
[plugins]
  required = ["idaes"]
  optional = []
[logging]
  version = 1
  disable_existing_loggers = false
  [logging.formatters.f1]
    format = "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
    datefmt = "%Y-%m-%d %H:%M:%S"
  [logging.handlers.console]
    class = "logging.StreamHandler"
    formatter = "f1"
    stream = "ext://sys.stdout"
  [logging.loggers.idaes]
    level = "INFO"
    propagate = true
    handlers = ["console"]
  [logging.loggers."idaes.init"]
    level = 5
    propagate = false
    handlers = ["console"]
  [logging.loggers."idaes.model"]
    level = "INFO"
    propagate = false
    handlers = ["console"]
"""

def new_idaes_config_block():
    _config = pyomo.common.config.ConfigBlock("idaes", implicit=False)
    _config.declare(
        "logging",
        pyomo.common.config.ConfigBlock(
            implicit=True,
            description="Logging configuration dictionary",
            doc="This stores the logging configuration. See the Python "
            "logging.config.dictConfig() documentation for details.",
        ),
    )
    _config.declare(
        "plugins",
        pyomo.common.config.ConfigBlock(
            implicit=False,
            description="Plugin search configuration",
            doc="Plugin search configuration",
        ),
    )
    _config.plugins.declare(
        "required",
        pyomo.common.config.ConfigValue(
            default=[],
            description="Modules with required plugins",
            doc="This is a string list of modules from which to load plugins. "
            "This will look in {module}.plugins for things to load. Exceptions"
            "raised while attempting to load these plugins are considered fatal. "
            "This is used for core plugins.",
        ),
    )
    _config.plugins.declare(
        "optional",
        pyomo.common.config.ConfigValue(
            default=[],
            description="Modules with optional plugins to load",
            doc="This is a string list of modules from which to load plugins. "
            "This will look in {module}.plugins for things to load. Exceptions "
            "raised while attempting to load these plugins will be logged but "
            "are nonfatal. This is used for contrib plugins.",
        ),
    )

    _config.declare(
        "use_idaes_solvers",
        pyomo.common.config.ConfigValue(
            default=True,
            description="Add the IDAES bin directory to the path.",
            doc="Add the IDAES bin directory to the path such that solvers provided "
            "by IDAES will be used in preference to previously installed solvers.",
        ),
    )

    _config.declare(
        "default_binary_url",
        pyomo.common.config.ConfigValue(
            default=None,
            description="URL from which to download binaries by default",
        ),
    )
    return _config


def read_config(read_config, write_config):
    """Read either a TOML formatted config file or a configuration dictionary.
    Args:
        config: A config file path or dict
    Returns:
        None
    """
    config_file = None
    if read_config is None:
        return
    elif isinstance(read_config, dict):
        pass  # don't worry this catches ConfigBlock too it seems
    else:
        config_file = read_config
        try:
            with open(config_file, "r") as f:
                write_config = toml.load(f)
        except IOError:  # don't require config file
            _log.debug("Config file {} not found (this is okay)".format(read_config))
            return
    write_config.set_value(read_config)
    logging.config.dictConfig(write_config["logging"])
    if config_file is not None:
        _log.debug("Read config {}".format(config_file))


def create_dir(d):
    """Create a directory if it doesn't exist.

    Args:
        d(str): directory path to create

    Retruns:
        None
    """
    if os.path.exists(d):
        return
    else:
        os.mkdir(d)

def import_packages(packages, optional=True):
    """Import plugin package, condensed from pyomo.environ.__init__.py
    Args:
        packages: list of packages in which to look for plugins
        optional: true, log ImportError but continue; false, raise if ImportError
    Returns:
        None
    """
    for name in packages:
        pname = name + '.plugins'  # look in plugins sub-package
        try:
            pkg = importlib.import_module(pname)
        except ImportError as e:
            _log.exception("failed to import plugin: {}".format(pname))
            if not optional:
                raise e
        if hasattr(pkg, 'load'):  # run load function for a module if it exists
            pkg.load()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Search through the code and index static information in the DMF.
"""
# stdlib
import glob
import importlib
import inspect
import logging
import os
import pprint
import re

__author__ = 'Dan Gunter <dkgunter@lbl.gov>'

_log = logging.getLogger(__name__)


class Walker(object):
    def walk(self, visitor):
        """Interface for walkers.

        Args:
            visitor (Visitor): Class whose `visit` method will be
                               called for each item.

        Returns:
            None
        """
        pass


class ModuleClassWalker(Walker):
    """Walk modules from a given root (e.g. 'idaes'), and visit
    all classes in those modules whose name matches a given pattern.

    Example usage::

        walker = ModuleClassWalker(from_pkg=idaes,
                                   class_expr='_PropertyParameter.*')

        walker.walk(PrintMetadataVisitor())  # see below

    """

    def __init__(self, from_path=None, from_pkg=None, class_expr=None,
                 parent_class=None, suppress_warnings=False,
                 exclude_testdirs=True, exclude_tests=True,
                 exclude_init=True, exclude_setup=True,
                 exclude_dirs=None):
        """Constructor. Create from either a path or package root, or both.
        If just one is given, the other is deduced. At least one must be
        given.

        Args:
            from_path (str): Path to start looking for modules. Overrides
                            'from_pkg'.
            from_pkg (module): Package from which to start looking for modules.
                            The path of the package will depend on the Python
                            environment and path.
            class_expr (str): Regular expression for the classes to find.
            parent_class (class): If given, filter for this parent class.
                                  Will match class_expr OR parent_class.
            suppress_warnings (bool): If True do not log any warning messages.
            exclude_testdirs (bool): If True (the default), exclude test dirs
            exclude_tests (bool): If True (the default), exclude test files
            exclude_init (bool): If True (the default), exclude __init___.py
            exclude_setup (bool): If True (the default), exclude setup.py
            exclude_dirs (list): List of directory (regex) patterns to exclude.
                                 These will be prefixed with a directory
                                 separator (e.g. '/') but otherwise can be
                                 any path expression.
        """
        if from_path:
            self._root = from_path
            if from_pkg:
                self._pkg = from_pkg.__name__
            else:
                self._pkg = str(os.path.basename(self._root))
        elif from_pkg:
            self._pkg = from_pkg.__name__
            try:
                self._root = from_pkg.__path__[0]
            except AttributeError:
                raise IOError('Root directory cannot be deduced from'
                              ' package "{}": no __path__ attribute.'
                              .format(self._pkg))
        else:
            raise ValueError('Missing arguments: either "from_pkg" or '
                             '"from_path" must be given')
        if not os.path.isdir(self._root):
            raise IOError('Root directory "{}"'.format(self._root))
        self._expr = re.compile(class_expr) if class_expr else None
        self._warn = not suppress_warnings
        self._parent = parent_class
        # build regular expression of things to exclude
        expr_list, psep = [], os.path.sep
        if exclude_testdirs:
            expr_list.append('{sl}tests?{sl}'.format(sl=psep))
        if exclude_tests:
            expr_list.append(r'tests?_.*.py')
        if exclude_init:
            expr_list.append(r'__init__\.py')
        if exclude_setup:
            expr_list.append(r'setup\.py')
        if exclude_dirs:
            for ed in exclude_dirs:
                expr_list.append('{sl}{d}'.format(sl=psep, d=ed))
        self._exclude_expr = re.compile('|'.join(expr_list))
        self._history = []
        # print('@@ exclude expr={}'.format(self._exclude_expr.pattern))

    def walk(self, visitor):
        modules = self._get_modules()
        self._visit_subclasses(modules, visitor.visit)

    def get_indexed_classes(self):
        return self._history

    def _get_modules(self):
        _log.debug('getting modules from root: {}'.format(self._root))
        # change file paths at 'root' to module paths from 'pkgroot'
        n, module_list = len(self._root), []
        for f in self._python_files():
            module_path = os.path.splitext(f[n + 1:])[0]
            module_path = module_path.replace(os.path.sep, '.')
            module_list.append(self._pkg + '.' + module_path)
        return module_list

    def _python_files(self):
        q = [self._root]
        while q:
            curdir = q.pop()
            for f in glob.glob(os.path.join(curdir, '*.py')):
                if not self._exclude_expr.search(f):
                    yield f
            for d in os.listdir(curdir):
                path = os.path.join(curdir, d)
                if os.path.isdir(path):
                    q.append(path)

    def _visit_subclasses(self, modules, visit):
        for modname in modules:
            _log.debug('visit module: {}'.format(modname))
            try:
                mod = importlib.import_module(modname)
            except Exception:
                if self._warn:
                    _log.warn('Error during import of module: {}. Ignoring.'
                              .format(modname))
                continue
            for item in dir(mod):
                x = getattr(mod, item)
                if inspect.isclass(x):
                    fullname = modname + '.' + x.__name__
                    if not self._expr and not self._parent:
                        if visit(x):
                            self._history.append(fullname)
                    elif self._expr:
                        if self._expr.match(x.__name__):
                            if visit(x):
                                self._history.append(fullname)
                    elif self._parent and issubclass(x, self._parent):
                        if visit(x):
                            self._history.append(fullname)


class Visitor(object):
    """Interface for the 'visitor' class passed to Walker subclasses'
    `walk()` method.
    """

    def visit(self, obj):
        """Visit one object.

        Args:
            obj (object): Some object to operate on.

        Returns:
            True if visit succeeded, else False
        """
        pass


class PropertyMetadataVisitor(Visitor):
    """Visit something implementing :class:`HasPropertyClassMetadata` and
    pass that metadata, as a dict, to the `visit_metadata()` method,
    which should be implemented by the subclass.
    """

    def visit(self, obj):
        """Visit one object.

        Args:
            obj (idaes.core.property_base.HasPropertyClassMetadata): The object

        Returns:
            True if visit succeeded, else False
        """
        result, meta = True, None
        try:
            meta = obj.get_metadata()
        except (AttributeError, TypeError) as err:
            module = obj.__module__
            _log.debug('Cannot get metadata for {}.{}: {}'
                       .format(module, obj.__name__, err))
            result = False
        except NotImplementedError:
            module = obj.__module__
#            if not module.startswith('idaes.core.'):
            _log.warn('{} in module "{}" does not define its metadata'
                      .format(obj.__name__, module))
            result = False
        if result:
            self.visit_metadata(obj, meta)
        return result

    def visit_metadata(self, obj, meta):
        """Do something with the metadata.

        Args:
            obj (object): Object from which metadata was pulled, for context.
            meta (idaes.core.property_base.PropertyClassMetadata): The metadata

        Returns:
            None
        """
        pass


class PrintPropertyMetadataVisitor(PropertyMetadataVisitor):
    def visit_metadata(self, obj, meta):
        """Print the module and class of the object, and
        then the metadata dict, to standard output.
        """
        meta_dict = {
            'properties': meta.properties,
            'default_units': meta.default_units
        }
        print('----')
        print('{}: class {}'.format(obj.__module__, obj.__name__))
        print('  metadata: {}'.format(pprint.pformat(meta_dict)))


# This is for testing


class _TestClass(object):
    pass


class _TestClass1(_TestClass):
    @classmethod
    def get_metadata(cls):
        return {'x': 1}


class _TestClass2(_TestClass):
    @classmethod
    def get_metadata(cls):
        return {'x': 2}


class _TestClass3(_TestClass):
    @classmethod
    def get_metadata(cls):
        return {'x': 3}



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Errors
"""
from idaes.surrogate.alamopy import writethis


def almerror(code):
    # reports errors encountered

    if (code == 'p1'):
        writethis('Inconsistent number of data points provided between input arrays')
        quit()
    elif (code == 'p2'):
        writethis('Error in structure of validation data')
        quit()
    elif (code == 'p3'):
        writethis('Inputs option not understood')
        quit()


class AlamoError(Exception):
    pass


class AlamoInputError(AlamoError):
    def __init__(self, msg):
        msg = 'Bad input to doalamo(): {}'.format(msg)
        super(AlamoInputError, self).__init__(msg)
